﻿/*
 * Copyright (C) 2010 Guillaume Duc <guillaume.duc@telecom-paristech.fr>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

using System;
namespace AttackWrapper
{
    class Program
    {
        private TraceDirectory traceDirectory;
        private AttackBase attack;
        private ResultFile resultFile;

        private byte[] correctSubkey;

        private string traceDirectoryname;
        private string resultFilename;
        private int numKeyToAttack;
        private int numTraces;

        public Program()
        {
            this.attack = new ReferenceAttack(); // The reference attack
            //this.attack = new Attack(); // Your attack (uncomment this line to use your attack)

            this.resultFile = new ResultFile();
            this.correctSubkey = new byte[Constants.TRACE_KEY_SIZE_BYTES];
        }

        private void parseCommandLine(string[] args)
        {
            // Parameters:
            // 1st: Key number
            // 2nd: Number of traces
            // 3rd: Result filename
            // 4th: Trace directory name

            if (args.Length < 4)
            {
                Console.WriteLine("Command line arguments:");
                Console.WriteLine("AttackWrapper.exe key_number number_traces result_filename trace_directory");
                Console.WriteLine("Example:");
                Console.WriteLine("AttackWrapper.exe 0 20000 result.bin traces");

                throw new ArgumentException();
            }

            numKeyToAttack = Convert.ToInt32(args[0]);
            numTraces = Convert.ToInt32(args[1]);
            resultFilename = args[2];
            traceDirectoryname = args[3];

            Console.WriteLine("Number of key to attack: {0}", numKeyToAttack);
            Console.WriteLine("Number of traces to use: {0}", numTraces);
            Console.WriteLine("Result filename: {0}", resultFilename);
            Console.WriteLine("Trace directory: {0}", traceDirectoryname);
        }

        private void init()
        {
            traceDirectory = new TraceDirectory(traceDirectoryname, numKeyToAttack);
            traceDirectory.parseDirectory();
            traceDirectory.findKey();

            AES.AES_key_schedule(traceDirectory.attackedKeyBytes, correctSubkey, attack.subkey_num);

            resultFile.subkeyNum = (uint)attack.subkey_num;
            resultFile.numIterations = (uint)numTraces;
            resultFile.numKey = (uint)numKeyToAttack;
            Array.Copy(traceDirectory.attackedKeyBytes, resultFile.correctKey, Constants.TRACE_KEY_SIZE_BYTES);
            Array.Copy(correctSubkey, resultFile.correctSubkey, Constants.TRACE_KEY_SIZE_BYTES);

            resultFile.openToWrite(resultFilename);
            resultFile.writeHeaders();
        }

        private void launchAttack()
        {
            Trace currentTrace;
            PartialResult partialResult;

            Console.WriteLine("Launching profiling phase...");

            attack.performProfilingPhase();

            Console.WriteLine("Launch attack phase...");

            for (int tracen = 0; tracen < numTraces; tracen++)
            {
                currentTrace = traceDirectory.getTrace(tracen);

                Console.Write("Trace #{0} (p={1},c={2})... ", tracen, 
                    Tools.byteArrayToString(currentTrace.plaintext),
                    Tools.byteArrayToString(currentTrace.ciphertext));

                DateTime startTime = DateTime.Now;

                partialResult = attack.attackPhase(currentTrace);

                DateTime endTime = DateTime.Now;
                TimeSpan duration = endTime - startTime;

                resultFile.writePartialResults(currentTrace, partialResult, duration.TotalSeconds);

                Console.WriteLine("done ({0:F2} s)", duration.TotalSeconds);
            }

            Console.WriteLine("Attack terminated");

            resultFile.close();
        }

        static void Main(string[] args)
        {
            Program prog = new Program();
            prog.parseCommandLine(args);
            prog.init();
            prog.launchAttack();
        }
    }
}
