﻿/*
 * Copyright (C) 2010 Guillaume Duc <guillaume.duc@telecom-paristech.fr>,
 *                    Sylvain Guilley <sylvain.guilley@telecom-paristech.fr>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * This is the reference attack for the DPA contest, developed initialy in C++
 * by Sylvain Guilley and adapted to C# by Guillaume Duc
 */

using System;
namespace AttackWrapper
{
    class ReferenceAttack : AttackBase
    {
        // The sum of traces (for the average trace)
        private float[] t;
        // The sum of the traces weighted by the selection function (for the covariance trace)
        private float[,,] tw;
        // The sum of weights (for the average trace)
        private float[,] w;
        // The sum of squared traces (for the variance trace)
        private float[] t2;
        // The sum of squared weights (for the variance of the weights)
        private float[,] w2;

        private Score[] best_score;

        private int n;


        public ReferenceAttack()
        {
            subkey_num = 10;

            t = new float[Constants.TRACE_NUM_POINTS];
            tw = new float[16,256,Constants.TRACE_NUM_POINTS];
            w = new float[16,256];
            t2 = new float[Constants.TRACE_NUM_POINTS];
            w2 = new float[16,256];
            best_score = new Score[256];
            n = 0;
        }

        // Perform the profiling phase of the attack (nothing if not a template attack)
        public override void performProfilingPhase()
        {

        }

        // Main attack: process a trace and return the result
        public override PartialResult attackPhase(Trace trace)
        {
            PartialResult partialResult = new PartialResult(subkey_num);

	        // 1. Updating the estimators:

	        n++; // Number of processed waves == number of calls of this function

	        for( int sample=0; sample<Constants.TRACE_NUM_POINTS; sample++)
	        {
		        t [sample] +=  (float) trace.samples[sample];
                t2[sample] += ((float) trace.samples[sample]) * ((float) trace.samples[sample]);
	        }

	        for( uint byteSK=0; byteSK<16; byteSK++)
	        {
		        // Let x be the position 0 <= x < 16.
		        // It is in column x/4 and in row x%4, because the state is like that:
		        //
		        // [  0  4  8 12 ]
		        // [  1  5  9 13 ]
		        // [  2  6 10 14 ]
		        // [  3  7 11 15 ]
		        //
		        // Therefore, an hypothesis on the key 0 <= k < 16 will lead to guessing
		        // a previous state (InvShiftRows) with the locations as:
		        //
		        // [  0  4  8 12 ]
		        // [ 13  1  5  9 ]
		        // [ 10 14  2  6 ]
		        // [  7 11 15  3 ]

		        // The differential waves:
		        float[,] cpa_wave = new float[256,Constants.TRACE_NUM_POINTS];

                Locus[] locus = new Locus[256];
                Locus the_locus = new Locus();

		        for( int kguess=0; kguess<256; kguess++ )
		        {
                    locus[kguess] = new Locus();
			        locus[kguess].sample = 0;
			        locus[kguess].cpa    = -1.0f; // Minimum possible value
		        }

		        for( uint kguess=0; kguess<256; kguess++ )
		        {
			        byte final_state   = trace.ciphertext[ get_previous_location( byteSK ) ];
				    byte initial_state = InvSubBytes[ trace.ciphertext[ byteSK ] ^ kguess ];

			        // Also: final_state = c ; initial state = InvSubBytes o InvShiftRows o InvAddRoundKey( c ),
			        // therefore a value of c XOR InvSubBytes o InvShiftRows o InvAddRoundKey( c ), which is equal to:
			        // ShiftRows( c ) XOR InvSubBytes o InvAddRoundKey( c ), because of the linearity of ShiftRows w.r.t. XOR and
			        // its commutatitivity with InvSubBytes (working on bytes vs with bytes).
			        uint sensitive_transition = hw( ((uint)initial_state) ^ ((uint)final_state) ); // The weight used in the linear prediction

			        w [byteSK,kguess] +=  (float)sensitive_transition;
			        w2[byteSK,kguess] += ((float)sensitive_transition)*((float)sensitive_transition);

			        for( int sample=0; sample<Constants.TRACE_NUM_POINTS; sample++)
			        {
				        tw[byteSK,kguess,sample] += (float) trace.samples[sample] * (float)sensitive_transition;
			        }


			        // 2. Computing the differential waves
			        float h_var = ( w2[byteSK,kguess] - w[byteSK,kguess] * w[byteSK,kguess] / n ) / n;
			        for( int sample=0; sample<Constants.TRACE_NUM_POINTS; sample++ )
			        {
				        // Covariance
				        cpa_wave[ kguess,sample ] = ( tw[byteSK,kguess,sample] - w [byteSK,kguess] * t[sample] / n) / n;

				        // Normalization by the trace
				        float t_var = ( t2[sample] - t[sample] * t[sample] / n ) / n;
				        if( t_var == 0.0f )
					        cpa_wave[ kguess,sample ] = 0.0f; // To avoid a division by zero ... we reset
				        else
        					cpa_wave[ kguess,sample ] /= (float) Math.Sqrt( t_var );

				        // Normalization by the weight
				        if( h_var == 0 )
        					cpa_wave[ kguess,sample ] = 0.0f; // To avoid a division by zero ... we reset
        				else
                			cpa_wave[ kguess,sample ] /= (float) Math.Sqrt( h_var );
			        }

			        // Finding the maximum for this guess
			        for( int sample=0; sample<Constants.TRACE_NUM_POINTS; sample++ )
			        {
				        // We do not consider CPA peaks in absolute value ;
				        // rather, they are in "relative" = "signed" values, i.e. compared plain.
				        if( cpa_wave[kguess,sample] > locus[kguess].cpa )
				        {
					        locus[kguess].sample = sample; // X
					        locus[kguess].cpa    = cpa_wave[kguess,sample]; // Y
				        }
			        }
		        }

		        // Looking for the time sample where the maximum amongst all key guesses is
		        the_locus.sample = 0;
		        the_locus.cpa    = -1.0f; // The minimum possible value
		        for( int kguess=0; kguess<256; kguess++ )
		        {
			        if( locus[kguess].cpa > the_locus.cpa )
			        {
				        the_locus.sample = locus[kguess].sample;
				        the_locus.cpa    = locus[kguess].cpa;
			        }
		        }

        		// Sorting
		        for( int kguess=0; kguess<256; kguess++ )
		        {
			        // All CPA points (for each key guess) are considered at the same date
                    best_score[kguess] = new Score((byte)kguess, cpa_wave[ kguess,the_locus.sample ]);
		        }

                Array.Sort(best_score);

		        for( int kguess=0; kguess<256; kguess++ )
		        {
                    // From the best CPA (highest == most probable key) to the poorest CPA
			        partialResult.bytes[byteSK,kguess] = best_score[kguess].kguess;
		        }

	        }

            return partialResult;
        }


        // ShiftRows
        private uint get_previous_location(uint x9)
        {
	        switch( x9 ) // and returns x10
	        {
		        case  0: return  0;
		        case  1: return  5;
		        case  2: return 10;
		        case  3: return 15;
		        case  4: return  4;
		        case  5: return  9;
		        case  6: return 14;
		        case  7: return  3;
		        case  8: return  8;
		        case  9: return 13;
		        case 10: return  2;
		        case 11: return  7;
		        case 12: return 12;
		        case 13: return  1;
		        case 14: return  6;
		        case 15: return 11;
	        }
            return 0;
        }
        
        private byte[] InvSubBytes = // Inverse of S
        {
	        0x52, 0x09, 0x6a, 0xd5, 0x30, 0x36, 0xa5, 0x38, 0xbf, 0x40, 0xa3, 0x9e, 0x81, 0xf3, 0xd7, 0xfb,
	        0x7c, 0xe3, 0x39, 0x82, 0x9b, 0x2f, 0xff, 0x87, 0x34, 0x8e, 0x43, 0x44, 0xc4, 0xde, 0xe9, 0xcb,
	        0x54, 0x7b, 0x94, 0x32, 0xa6, 0xc2, 0x23, 0x3d, 0xee, 0x4c, 0x95, 0x0b, 0x42, 0xfa, 0xc3, 0x4e,
	        0x08, 0x2e, 0xa1, 0x66, 0x28, 0xd9, 0x24, 0xb2, 0x76, 0x5b, 0xa2, 0x49, 0x6d, 0x8b, 0xd1, 0x25,
	        0x72, 0xf8, 0xf6, 0x64, 0x86, 0x68, 0x98, 0x16, 0xd4, 0xa4, 0x5c, 0xcc, 0x5d, 0x65, 0xb6, 0x92,
	        0x6c, 0x70, 0x48, 0x50, 0xfd, 0xed, 0xb9, 0xda, 0x5e, 0x15, 0x46, 0x57, 0xa7, 0x8d, 0x9d, 0x84,
	        0x90, 0xd8, 0xab, 0x00, 0x8c, 0xbc, 0xd3, 0x0a, 0xf7, 0xe4, 0x58, 0x05, 0xb8, 0xb3, 0x45, 0x06,
	        0xd0, 0x2c, 0x1e, 0x8f, 0xca, 0x3f, 0x0f, 0x02, 0xc1, 0xaf, 0xbd, 0x03, 0x01, 0x13, 0x8a, 0x6b,
	        0x3a, 0x91, 0x11, 0x41, 0x4f, 0x67, 0xdc, 0xea, 0x97, 0xf2, 0xcf, 0xce, 0xf0, 0xb4, 0xe6, 0x73,
	        0x96, 0xac, 0x74, 0x22, 0xe7, 0xad, 0x35, 0x85, 0xe2, 0xf9, 0x37, 0xe8, 0x1c, 0x75, 0xdf, 0x6e,
	        0x47, 0xf1, 0x1a, 0x71, 0x1d, 0x29, 0xc5, 0x89, 0x6f, 0xb7, 0x62, 0x0e, 0xaa, 0x18, 0xbe, 0x1b,
	        0xfc, 0x56, 0x3e, 0x4b, 0xc6, 0xd2, 0x79, 0x20, 0x9a, 0xdb, 0xc0, 0xfe, 0x78, 0xcd, 0x5a, 0xf4,
	        0x1f, 0xdd, 0xa8, 0x33, 0x88, 0x07, 0xc7, 0x31, 0xb1, 0x12, 0x10, 0x59, 0x27, 0x80, 0xec, 0x5f,
	        0x60, 0x51, 0x7f, 0xa9, 0x19, 0xb5, 0x4a, 0x0d, 0x2d, 0xe5, 0x7a, 0x9f, 0x93, 0xc9, 0x9c, 0xef,
	        0xa0, 0xe0, 0x3b, 0x4d, 0xae, 0x2a, 0xf5, 0xb0, 0xc8, 0xeb, 0xbb, 0x3c, 0x83, 0x53, 0x99, 0x61,
	        0x17, 0x2b, 0x04, 0x7e, 0xba, 0x77, 0xd6, 0x26, 0xe1, 0x69, 0x14, 0x63, 0x55, 0x21, 0x0c, 0x7d
        };

        // The Hamming weight, aka hw, of a byte, unsigned int, size_t, etc. (ancillary function for parity)
        private uint hw(uint bitvector)
        {
            uint w = 0; // Initial weight is set to zero
            while(bitvector != 0u)
            {
                w += bitvector & 0x01u; // Bits accumulation
                bitvector >>= 1; // Unsignedness is necessary here to avoid sign extension
            }
            return w;
        }


        class Score : IComparable
        {
            internal byte kguess;
	        internal float cpa;

            internal Score(byte kguess, float cpa)
            {
                this.kguess = kguess;
                this.cpa = cpa;
            }

            public int CompareTo(object obj)
            {
                if (obj is Score)
                {
                    Score sc = (Score)obj;

                    if (this.cpa > sc.cpa) return -1;
                    if (this.cpa == sc.cpa) return this.cpa.CompareTo(sc.cpa);
                    return 1;
                }

                throw new ArgumentException();
            }
        }

        class Locus
        {
            internal int sample;
            internal float cpa;
        }
    }
}
