﻿/*
 * Copyright (C) 2010 Guillaume Duc <guillaume.duc@telecom-paristech.fr>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

using System;
using System.Collections.Generic;

using AttackWrapper;
using System.IO;
using System.Text;

namespace ComputeResults
{
    class Program
    {
        private float[,] partialSuccessRate;
        private float[,] partialGuessingEntropy;
        private float[] globalSuccessRate;

        private uint[,] partialSuccessRateAcc;
        private uint[,] partialGuessingEntropyAcc;
        private uint[] globalSuccessRateAcc;

        private uint numIterations;
        private int numExperiments;

        private List<string> filenames;

        public Program()
        {
            numIterations = 0;
            numExperiments = 0;
            filenames = new List<string>();
        }

        public void parseCommandLine(string[] args)
        {
            if (args.Length == 0)
            {
                Console.WriteLine("Command line arguments:");
                Console.WriteLine("AttackWrapper.exe result1.bin result2.bin ...");

                throw new ArgumentException();
            }

            foreach (string filename in args)
            {
                filenames.Add(filename);
            }
        }

        public void createDataStructures()
        {
            partialSuccessRate = new float[Constants.TRACE_KEY_SIZE_BYTES, numIterations];
            partialGuessingEntropy = new float[Constants.TRACE_KEY_SIZE_BYTES, numIterations];
            globalSuccessRate = new float[numIterations];

            partialSuccessRateAcc = new uint[Constants.TRACE_KEY_SIZE_BYTES, numIterations];
            partialGuessingEntropyAcc = new uint[Constants.TRACE_KEY_SIZE_BYTES, numIterations];
            globalSuccessRateAcc = new uint[numIterations];
        }

        public void accumulateResults()
        {
            foreach (string filename in filenames)
            {
                Console.WriteLine("Processing file {0}...", filename);

                ResultFile resultFile = new ResultFile();

                resultFile.openToRead(filename);
                resultFile.readHeaders();

                Console.WriteLine("Correct key = {0}", Tools.byteArrayToString(resultFile.correctSubkey));

                numExperiments++;

                if (numIterations == 0)
                {
                    numIterations = resultFile.numIterations;
                    createDataStructures();
                }

                for (int iteration = 0; iteration < numIterations; iteration++)
                {
                    PartialResult partialResult = resultFile.readPartialResults();

                    int numCorrectSubkeyBytes = 0;

                    for (int subkeyByte = 0; subkeyByte < Constants.TRACE_KEY_SIZE_BYTES; subkeyByte++)
                    {
                        uint byteIndex;

                        for (byteIndex = 0; byteIndex < 256; byteIndex++)
                        {
                            if (partialResult.bytes[subkeyByte, byteIndex] == resultFile.correctSubkey[subkeyByte])
                            {
                                partialGuessingEntropyAcc[subkeyByte, iteration] += byteIndex +  1;
                                break;
                            }
                        }

                        if (byteIndex == 256)
                            partialGuessingEntropyAcc[subkeyByte, iteration] += 256;

                        if (partialResult.bytes[subkeyByte, 0] == resultFile.correctSubkey[subkeyByte])
                        {
                            partialSuccessRateAcc[subkeyByte, iteration] += 1;
                            numCorrectSubkeyBytes++;
                        }
                    }

                    if (numCorrectSubkeyBytes == Constants.TRACE_KEY_SIZE_BYTES)
                        globalSuccessRateAcc[iteration]++;
                }
            }
        }

        public void computeResults()
        {
            for (int iteration = 0; iteration < numIterations; iteration++)
            {
                for (int subkeyByte = 0; subkeyByte < Constants.TRACE_KEY_SIZE_BYTES; subkeyByte++)
                {
                    partialGuessingEntropy[subkeyByte, iteration] = ((float)partialGuessingEntropyAcc[subkeyByte, iteration]) /
                        ((float)numExperiments);

                    partialSuccessRate[subkeyByte, iteration] = ((float)partialSuccessRateAcc[subkeyByte, iteration]) /
                        ((float)numExperiments);
                }

                globalSuccessRate[iteration] = ((float)globalSuccessRateAcc[iteration]) / ((float)numExperiments);
            }
        }

        public void dumpResults()
        {
            StreamWriter[] partialSuccessRateFiles = new StreamWriter[Constants.TRACE_KEY_SIZE_BYTES];
            StreamWriter[] partialGuessingEntropyFiles = new StreamWriter[Constants.TRACE_KEY_SIZE_BYTES];

            // Open the files
            StreamWriter globalSuccessRateFile = new StreamWriter(File.Open("global_success_rate.dat", FileMode.Create, FileAccess.Write),
                Encoding.ASCII);
            StreamWriter partialSuccessRateAllFile = new StreamWriter(File.Open("partial_success_rate_all.dat",
                FileMode.Create, FileAccess.Write), Encoding.ASCII);
            StreamWriter partialGuessingEntropyAllFile = new StreamWriter(File.Open("partial_guessing_entropy_all.dat",
                FileMode.Create, FileAccess.Write), Encoding.ASCII);

            for (int subkeyByte = 0; subkeyByte < Constants.TRACE_KEY_SIZE_BYTES; subkeyByte++)
            {
                partialGuessingEntropyFiles[subkeyByte] = new StreamWriter(File.Open("partial_guessing_entropy_b="+subkeyByte+".dat",
                    FileMode.Create, FileAccess.Write), Encoding.ASCII);
                partialSuccessRateFiles[subkeyByte] = new StreamWriter(File.Open("partial_success_rate_b=" + subkeyByte + ".dat",
                    FileMode.Create, FileAccess.Write), Encoding.ASCII);
            }


            // Write headers
            globalSuccessRateFile.WriteLine("# Trace_num Global_success_rate");
            partialSuccessRateAllFile.WriteLine("# Trace_num Subkey_byte Partial_success_rate");
            partialGuessingEntropyAllFile.WriteLine("# Trace_num Subkey_byte Partial_guessing_entropy");
            for (int subkeyByte = 0; subkeyByte < Constants.TRACE_KEY_SIZE_BYTES; subkeyByte++)
            {
                partialGuessingEntropyFiles[subkeyByte].WriteLine("# Trace_num Partial_guessing_entropy");
                partialSuccessRateFiles[subkeyByte].WriteLine("# Trace_num Partial_success_rate");
            }


            // Output the results
            for (int iteration = 0; iteration < numIterations; iteration++)
            {
                for (int subkeyByte = 0; subkeyByte < Constants.TRACE_KEY_SIZE_BYTES; subkeyByte++)
                {
                    partialGuessingEntropyFiles[subkeyByte].WriteLine("{0:D} {1:F}", iteration,
                        partialGuessingEntropy[subkeyByte,iteration]);
                    partialSuccessRateFiles[subkeyByte].WriteLine("{0:D} {1:F}", iteration,
                        partialSuccessRate[subkeyByte, iteration]);

                    partialGuessingEntropyAllFile.WriteLine("{0:D} {1:D} {2:F}", iteration, subkeyByte,
                        partialGuessingEntropy[subkeyByte, iteration]);
                    partialSuccessRateAllFile.WriteLine("{0:D} {1:D} {2:F}", iteration, subkeyByte,
                        partialSuccessRate[subkeyByte, iteration]);
                }

                globalSuccessRateFile.WriteLine("{0:D} {1:F}", iteration, globalSuccessRate[iteration]);

                partialGuessingEntropyAllFile.WriteLine();
                partialSuccessRateAllFile.WriteLine();
            }


            // Close the files
            globalSuccessRateFile.Close();
            partialSuccessRateAllFile.Close();
            partialGuessingEntropyAllFile.Close();

            for (int subkeyByte = 0; subkeyByte < Constants.TRACE_KEY_SIZE_BYTES; subkeyByte++)
            {
                partialGuessingEntropyFiles[subkeyByte].Close();
                partialSuccessRateFiles[subkeyByte].Close();
            }
        }

        static void Main(string[] args)
        {
            Program program = new Program();
            program.parseCommandLine(args);
            program.accumulateResults();
            program.computeResults();
            program.dumpResults();
        }
    }
}
